@@ -44,21 +44,22 @@ MVM_PUBLIC void MVM_repr_pos_set_elems(MVMThreadContext *tc, MVMObject *obj, MVM
         OBJECT_BODY(obj), elems);
 }
 
-static void int_array_to_c_array(MVMThreadContext *tc, MVMObject *arr, MVMint64 *elems, MVMint64 **values) {
+static void populate_indices_array(MVMThreadContext *tc, MVMObject *arr, MVMint64 *elems) {
     MVMint64 i;
     *elems = MVM_repr_elems(tc, arr);
-    *values = *elems ? MVM_malloc(*elems * sizeof(MVMint64)) : NULL;
+    if (*elems > tc->num_multi_dim_indices)
+        tc->multi_dim_indices = MVM_realloc(tc->multi_dim_indices,
+            *elems * sizeof(MVMint64));
     for (i = 0; i < *elems; i++)
-        (*values)[i] = MVM_repr_at_pos_i(tc, arr, i);
+        tc->multi_dim_indices[i] = MVM_repr_at_pos_i(tc, arr, i);
 }
 
 void MVM_repr_set_dimensions(MVMThreadContext *tc, MVMObject *obj, MVMObject *dims) {
     if (IS_CONCRETE(obj)) {
-        MVMint64 num_dims, *c_dims;
-        int_array_to_c_array(tc, dims, &num_dims, &c_dims);
+        MVMint64 num_dims;
+        populate_indices_array(tc, dims, &num_dims);
         REPR(obj)->pos_funcs.set_dimensions(tc, STABLE(obj), obj,
-            OBJECT_BODY(obj), num_dims, c_dims);
-        MVM_free(c_dims);
+            OBJECT_BODY(obj), num_dims, tc->multi_dim_indices);
     }
     else {
         MVM_exception_throw_adhoc(tc, "Cannot set dimensions on a type object");
@@ -110,11 +111,10 @@ MVMObject * MVM_repr_at_pos_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx
 }
 
 static void at_pos_multidim(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMRegister *value, MVMuint16 kind) {
-    MVMint64 num_indices, *c_indices;
-    int_array_to_c_array(tc, indices, &num_indices, &c_indices);
+    MVMint64 num_indices;
+    populate_indices_array(tc, indices, &num_indices);
     REPR(obj)->pos_funcs.at_pos_multidim(tc, STABLE(obj), obj,
-        OBJECT_BODY(obj), num_indices, c_indices, value, kind);
-    MVM_free(c_indices);
+        OBJECT_BODY(obj), num_indices, tc->multi_dim_indices, value, kind);
 }
 
 MVMint64 MVM_repr_at_pos_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
@@ -230,11 +230,10 @@ void MVM_repr_bind_pos_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVM
 }
 
 static void bind_pos_multidim(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMRegister value, MVMuint16 kind) {
-    MVMint64 num_indices, *c_indices;
-    int_array_to_c_array(tc, indices, &num_indices, &c_indices);
+    MVMint64 num_indices;
+    populate_indices_array(tc, indices, &num_indices);
     REPR(obj)->pos_funcs.bind_pos_multidim(tc, STABLE(obj), obj,
-        OBJECT_BODY(obj), num_indices, c_indices, value, kind);
-    MVM_free(c_indices);
+        OBJECT_BODY(obj), num_indices, tc->multi_dim_indices, value, kind);
 }
 
 void MVM_repr_bind_pos_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMint64 value) {
@@ -544,6 +543,12 @@ MVMString * MVM_repr_get_str(MVMThreadContext *tc, MVMObject *obj) {
     return REPR(obj)->box_funcs.get_str(tc, STABLE(obj), obj, OBJECT_BODY(obj));
 }
 
+MVMuint64 MVM_repr_get_uint(MVMThreadContext *tc, MVMObject *obj) {
+    if (!IS_CONCRETE(obj))
+        MVM_exception_throw_adhoc(tc, "Cannot unbox a type object");
+    return REPR(obj)->box_funcs.get_uint(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+}
+
 void MVM_repr_set_int(MVMThreadContext *tc, MVMObject *obj, MVMint64 val) {
     REPR(obj)->box_funcs.set_int(tc, STABLE(obj), obj, OBJECT_BODY(obj), val);
 }
@@ -556,6 +561,10 @@ void MVM_repr_set_str(MVMThreadContext *tc, MVMObject *obj, MVMString *val) {
     REPR(obj)->box_funcs.set_str(tc, STABLE(obj), obj, OBJECT_BODY(obj), val);
 }
 
+void MVM_repr_set_uint(MVMThreadContext *tc, MVMObject *obj, MVMuint64 val) {
+    REPR(obj)->box_funcs.set_uint(tc, STABLE(obj), obj, OBJECT_BODY(obj), val);
+}
+
 MVMObject * MVM_repr_box_int(MVMThreadContext *tc, MVMObject *type, MVMint64 val) {
     MVMObject *res;
     res = MVM_intcache_get(tc, type, val);
@@ -581,6 +590,12 @@ MVMObject * MVM_repr_box_str(MVMThreadContext *tc, MVMObject *type, MVMString *v
     return res;
 }
 
+MVMObject * MVM_repr_box_uint(MVMThreadContext *tc, MVMObject *type, MVMuint64 val) {
+    MVMObject *res = MVM_repr_alloc_init(tc, type);
+    MVM_repr_set_uint(tc, res, val);
+    return res;
+}
+
 MVM_PUBLIC MVMint64 MVM_repr_get_attr_i(MVMThreadContext *tc, MVMObject *object, MVMObject *type,
                                            MVMString *name, MVMint16 hint) {
     MVMRegister result_reg;
@@ -641,6 +656,15 @@ MVM_PUBLIC void MVM_repr_bind_attr_inso(MVMThreadContext *tc, MVMObject *object,
     MVM_SC_WB_OBJ(tc, object);
 }
 
+MVM_PUBLIC MVMint64 MVM_repr_attribute_inited(MVMThreadContext *tc, MVMObject *obj, MVMObject *type,
+                                              MVMString *name) {
+    if (!IS_CONCRETE(obj))
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a type object");
+    return REPR(obj)->attr_funcs.is_attribute_initialized(tc,
+        STABLE(obj), OBJECT_BODY(obj),
+        type, name, MVM_NO_HINT);
+}
+
 MVM_PUBLIC MVMint64    MVM_repr_compare_repr_id(MVMThreadContext *tc, MVMObject *object, MVMuint32 REPRId) {
     return object && REPR(object)->ID == REPRId ? 1 : 0;
 }
