@@ -31,20 +31,20 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
 static void serialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
     if (repr_data) {
-        MVM_serialization_write_varint(tc, writer, repr_data->primitive_type);
-        MVM_serialization_write_varint(tc, writer, repr_data->ref_kind);
+        MVM_serialization_write_int(tc, writer, repr_data->primitive_type);
+        MVM_serialization_write_int(tc, writer, repr_data->ref_kind);
     }
     else {
-        MVM_serialization_write_varint(tc, writer, 0);
-        MVM_serialization_write_varint(tc, writer, 0);
+        MVM_serialization_write_int(tc, writer, 0);
+        MVM_serialization_write_int(tc, writer, 0);
     }
 }
 
 /* Deserializes REPR data. */
 static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerializationReader *reader) {
     MVMNativeRefREPRData *repr_data = MVM_malloc(sizeof(MVMNativeRefREPRData));
-    repr_data->primitive_type = MVM_serialization_read_varint(tc, reader);
-    repr_data->ref_kind       = MVM_serialization_read_varint(tc, reader);
+    repr_data->primitive_type = MVM_serialization_read_int(tc, reader);
+    repr_data->ref_kind       = MVM_serialization_read_int(tc, reader);
     st->REPR_data = repr_data;
 }
 
@@ -54,7 +54,7 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
     switch (repr_data->ref_kind) {
         case MVM_NATIVEREF_REG_OR_LEX:
-            MVM_gc_worklist_add_frame(tc, worklist, ref->u.reg_or_lex.frame);
+            MVM_gc_worklist_add(tc, worklist, &ref->u.reg_or_lex.frame);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             MVM_gc_worklist_add(tc, worklist, &ref->u.attribute.obj);
@@ -64,17 +64,13 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
         case MVM_NATIVEREF_POSITIONAL:
             MVM_gc_worklist_add(tc, worklist, &ref->u.positional.obj);
             break;
+        case MVM_NATIVEREF_MULTIDIM:
+            MVM_gc_worklist_add(tc, worklist, &ref->u.multidim.obj);
+            MVM_gc_worklist_add(tc, worklist, &ref->u.multidim.indices);
+            break;
     }
 }
 
-/* Called by the VM in order to free memory associated with this object. */
-static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
-    MVMNativeRef *ref = (MVMNativeRef *)obj;
-    MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(obj)->REPR_data;
-    if (repr_data->ref_kind == MVM_NATIVEREF_REG_OR_LEX)
-        MVM_frame_dec_ref(tc, ref->body.u.reg_or_lex.frame);
-}
-
 /* Frees the representation data, if any. */
 static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
     MVM_free(st->REPR_data);
@@ -115,6 +111,9 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
                 else if (MVM_string_equal(tc, refkind_s, str_consts->positional)) {
                     kind = MVM_NATIVEREF_POSITIONAL;
                 }
+                else if (MVM_string_equal(tc, refkind_s, str_consts->multidim)) {
+                    kind = MVM_NATIVEREF_MULTIDIM;
+                }
                 else {
                     MVM_exception_throw_adhoc(tc, "NativeRef: invalid refkind in compose");
                 }
@@ -142,6 +141,44 @@ const MVMREPROps * MVMNativeRef_initialize(MVMThreadContext *tc) {
     return &this_repr;
 }
 
+static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMNativeRefREPRData * repr_data = (MVMNativeRefREPRData *)st->REPR_data;
+    MVMuint16              opcode    = ins->info->opcode;
+
+    if (!repr_data)
+        return;
+
+    if (repr_data->ref_kind != MVM_NATIVEREF_REG_OR_LEX)
+        return; /* TODO implement spesh for attribute and positional references */
+
+    switch (opcode) {
+        case MVM_OP_assign_i: {
+            MVMSpeshOperand target   = ins->operands[0];
+            MVMSpeshOperand value    = ins->operands[1];
+            if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_INT)
+                return; /* Shouldn't happen. so maybe throw an error? */
+            ins->info = MVM_op_get_op(MVM_OP_sp_deref_bind_i64);
+            ins->operands            = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+            ins->operands[0]         = target;
+            ins->operands[1]         = value;
+            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.reg_or_lex.var);
+            break;
+        }
+        case MVM_OP_decont_i: {
+            MVMSpeshOperand target   = ins->operands[0];
+            MVMSpeshOperand source   = ins->operands[1];
+            if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_INT)
+                return; /* Shouldn't happen. so maybe throw an error? */
+            ins->info = MVM_op_get_op(MVM_OP_sp_deref_get_i64);
+            ins->operands            = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+            ins->operands[0]         = target;
+            ins->operands[1]         = source;
+            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.reg_or_lex.var);
+            break;
+        }
+    }
+}
+
 static const MVMREPROps this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
@@ -160,15 +197,16 @@ static const MVMREPROps this_repr = {
     deserialize_repr_data,
     deserialize_stable_size,
     gc_mark,
-    gc_free,
+    NULL, /* gc_free */
     NULL, /* gc_cleanup */
     NULL, /* gc_mark_repr_data */
     gc_free_repr_data,
     compose,
-    NULL, /* spesh */
+    spesh, /* spesh */
     "NativeRef", /* name */
     MVM_REPR_ID_NativeRef,
-    1, /* refs_frames */
+    NULL, /* unmanaged_size */
+    NULL, /* describe_refs */
 };
 
 /* Validates the given type is a native reference of the required primitive
@@ -190,25 +228,31 @@ void MVM_nativeref_ensure(MVMThreadContext *tc, MVMObject *type, MVMuint16 wantp
 
 /* Creation of native references for registers. */
 static MVMObject * reg_or_lex_ref(MVMThreadContext *tc, MVMObject *type, MVMFrame *f, MVMRegister *r, MVMuint16 reg_type) {
-    MVMNativeRef *ref = (MVMNativeRef *)MVM_gc_allocate_object(tc, STABLE(type));
-    ref->body.u.reg_or_lex.frame = MVM_frame_inc_ref(tc, f);
+    MVMNativeRef *ref;
+    MVMROOT(tc, f, {
+        ref = (MVMNativeRef *)MVM_gc_allocate_object(tc, STABLE(type));
+    });
+    MVM_ASSIGN_REF(tc, &(ref->common.header), ref->body.u.reg_or_lex.frame, f);
     ref->body.u.reg_or_lex.var   = r;
     ref->body.u.reg_or_lex.type  = reg_type;
     return (MVMObject *)ref;
 }
-MVMObject * MVM_nativeref_reg_i(MVMThreadContext *tc, MVMFrame *f, MVMRegister *r) {
+MVMObject * MVM_nativeref_reg_i(MVMThreadContext *tc, MVMRegister *r) {
+    MVMFrame *f = MVM_frame_force_to_heap(tc, tc->cur_frame);
     MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
     if (ref_type)
         return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_int64);
     MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
 }
-MVMObject * MVM_nativeref_reg_n(MVMThreadContext *tc, MVMFrame *f, MVMRegister *r) {
+MVMObject * MVM_nativeref_reg_n(MVMThreadContext *tc, MVMRegister *r) {
+    MVMFrame *f = MVM_frame_force_to_heap(tc, tc->cur_frame);
     MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
     if (ref_type)
         return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_num64);
     MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
 }
-MVMObject * MVM_nativeref_reg_s(MVMThreadContext *tc, MVMFrame *f, MVMRegister *r) {
+MVMObject * MVM_nativeref_reg_s(MVMThreadContext *tc, MVMRegister *r) {
+    MVMFrame *f = MVM_frame_force_to_heap(tc, tc->cur_frame);
     MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
     if (ref_type)
         return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_str);
@@ -227,7 +271,9 @@ static MVMFrame * get_lexical_outer(MVMThreadContext *tc, MVMuint16 outers) {
     return f;
 }
 MVMObject * MVM_nativeref_lex_i(MVMThreadContext *tc, MVMuint16 outers, MVMuint16 idx) {
-    MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
+    MVMObject *ref_type;
+    MVM_frame_force_to_heap(tc, tc->cur_frame);
+    ref_type = MVM_hll_current(tc)->int_lex_ref;
     if (ref_type) {
         MVMFrame  *f = get_lexical_outer(tc, outers);
         MVMuint16 *lexical_types = f->spesh_cand && f->spesh_cand->lexical_types
@@ -235,14 +281,18 @@ MVMObject * MVM_nativeref_lex_i(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
             : f->static_info->body.lexical_types;
         MVMuint16 type = lexical_types[idx];
         if (type != MVM_reg_int64 && type != MVM_reg_int32 &&
-                type != MVM_reg_int16 && type != MVM_reg_int8)
+                type != MVM_reg_int16 && type != MVM_reg_int8 &&
+                type != MVM_reg_uint64 && type != MVM_reg_uint32 &&
+                type != MVM_reg_uint16 && type != MVM_reg_uint8)
             MVM_exception_throw_adhoc(tc, "getlexref_i: lexical is not an int");
         return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), type);
     }
     MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_n(MVMThreadContext *tc, MVMuint16 outers, MVMuint16 idx) {
-    MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
+    MVMObject *ref_type;
+    MVM_frame_force_to_heap(tc, tc->cur_frame);
+    ref_type = MVM_hll_current(tc)->num_lex_ref;
     if (ref_type) {
         MVMFrame  *f = get_lexical_outer(tc, outers);
         MVMuint16 *lexical_types = f->spesh_cand && f->spesh_cand->lexical_types
@@ -256,7 +306,9 @@ MVMObject * MVM_nativeref_lex_n(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
     MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_s(MVMThreadContext *tc, MVMuint16 outers, MVMuint16 idx) {
-    MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
+    MVMObject *ref_type;
+    MVM_frame_force_to_heap(tc, tc->cur_frame);
+    ref_type = MVM_hll_current(tc)->str_lex_ref;
     if (ref_type) {
         MVMFrame  *f = get_lexical_outer(tc, outers);
         MVMuint16 *lexical_types = f->spesh_cand && f->spesh_cand->lexical_types
@@ -299,19 +351,31 @@ static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMStri
     }
 }
 MVMObject * MVM_nativeref_lex_name_i(MVMThreadContext *tc, MVMString *name) {
-    MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
+    MVMObject *ref_type;
+    MVMROOT(tc, name, {
+        MVM_frame_force_to_heap(tc, tc->cur_frame);
+    });
+    ref_type = MVM_hll_current(tc)->int_lex_ref;
     if (ref_type)
         return lexref_by_name(tc, ref_type, name, MVM_reg_int64);
     MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_name_n(MVMThreadContext *tc, MVMString *name) {
-    MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
+    MVMObject *ref_type;
+    MVMROOT(tc, name, {
+        MVM_frame_force_to_heap(tc, tc->cur_frame);
+    });
+    ref_type = MVM_hll_current(tc)->num_lex_ref;
     if (ref_type)
         return lexref_by_name(tc, ref_type, name, MVM_reg_num64);
     MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
 }
 MVMObject * MVM_nativeref_lex_name_s(MVMThreadContext *tc, MVMString *name) {
-    MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
+    MVMObject *ref_type;
+    MVMROOT(tc, name, {
+        MVM_frame_force_to_heap(tc, tc->cur_frame);
+    });
+    ref_type = MVM_hll_current(tc)->str_lex_ref;
     if (ref_type)
         return lexref_by_name(tc, ref_type, name, MVM_reg_str);
     MVM_exception_throw_adhoc(tc, "No str lexical reference type registered for current HLL");
@@ -380,6 +444,37 @@ MVMObject * MVM_nativeref_pos_s(MVMThreadContext *tc, MVMObject *obj, MVMint64 i
     MVM_exception_throw_adhoc(tc, "No str positional reference type registered for current HLL");
 }
 
+/* Creation of native references for multi-dimensional positionals. */
+static MVMObject * md_posref(MVMThreadContext *tc, MVMObject *type, MVMObject *obj, MVMObject *indices) {
+    MVMNativeRef *ref;
+    MVMROOT(tc, obj, {
+    MVMROOT(tc, indices, {
+        ref = (MVMNativeRef *)MVM_gc_allocate_object(tc, STABLE(type));
+        MVM_ASSIGN_REF(tc, &(ref->common.header), ref->body.u.multidim.obj, obj);
+        MVM_ASSIGN_REF(tc, &(ref->common.header), ref->body.u.multidim.indices, indices);
+    });
+    });
+    return (MVMObject *)ref;
+}
+MVMObject * MVM_nativeref_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMObject *ref_type = MVM_hll_current(tc)->int_multidim_ref;
+    if (ref_type)
+        return md_posref(tc, ref_type, obj, indices);
+    MVM_exception_throw_adhoc(tc, "No int positional reference type registered for current HLL");
+}
+MVMObject * MVM_nativeref_multidim_n(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMObject *ref_type = MVM_hll_current(tc)->num_multidim_ref;
+    if (ref_type)
+        return md_posref(tc, ref_type, obj, indices);
+    MVM_exception_throw_adhoc(tc, "No num positional reference type registered for current HLL");
+}
+MVMObject * MVM_nativeref_multidim_s(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMObject *ref_type = MVM_hll_current(tc)->str_multidim_ref;
+    if (ref_type)
+        return md_posref(tc, ref_type, obj, indices);
+    MVM_exception_throw_adhoc(tc, "No str positional reference type registered for current HLL");
+}
+
 /* Reference read functions. These do no checks that the reference is of the
  * right kind and primitive type, they just go ahead and do the read. Thus
  * they are more suited to calling from optimized code. The checking path is
@@ -438,6 +533,18 @@ MVMString * MVM_nativeref_read_positional_s(MVMThreadContext *tc, MVMObject *ref
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
     return MVM_repr_at_pos_s(tc, ref->body.u.positional.obj, ref->body.u.positional.idx);
 }
+MVMint64 MVM_nativeref_read_multidim_i(MVMThreadContext *tc, MVMObject *ref_obj) {
+    MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
+    return MVM_repr_at_pos_multidim_i(tc, ref->body.u.multidim.obj, ref->body.u.multidim.indices);
+}
+MVMnum64 MVM_nativeref_read_multidim_n(MVMThreadContext *tc, MVMObject *ref_obj) {
+    MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
+    return MVM_repr_at_pos_multidim_n(tc, ref->body.u.multidim.obj, ref->body.u.multidim.indices);
+}
+MVMString * MVM_nativeref_read_multidim_s(MVMThreadContext *tc, MVMObject *ref_obj) {
+    MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
+    return MVM_repr_at_pos_multidim_s(tc, ref->body.u.multidim.obj, ref->body.u.multidim.indices);
+}
 
 /* Reference write functions. Same (non-checking) rules as the reads above. */
 void MVM_nativeref_write_reg_or_lex_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
@@ -470,7 +577,8 @@ void MVM_nativeref_write_reg_or_lex_n(MVMThreadContext *tc, MVMObject *ref_obj,
 }
 void MVM_nativeref_write_reg_or_lex_s(MVMThreadContext *tc, MVMObject *ref_obj, MVMString *value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    ref->body.u.reg_or_lex.var->s = value;
+    MVM_ASSIGN_REF(tc, &(ref->body.u.reg_or_lex.frame->header),
+        ref->body.u.reg_or_lex.var->s, value);
 }
 void MVM_nativeref_write_attribute_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
@@ -505,3 +613,16 @@ void MVM_nativeref_write_positional_s(MVMThreadContext *tc, MVMObject *ref_obj,
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
     MVM_repr_bind_pos_s(tc, ref->body.u.positional.obj, ref->body.u.positional.idx, value);
 }
+
+void MVM_nativeref_write_multidim_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
+    MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
+    MVM_repr_bind_pos_multidim_i(tc, ref->body.u.multidim.obj, ref->body.u.multidim.indices, value);
+}
+void MVM_nativeref_write_multidim_n(MVMThreadContext *tc, MVMObject *ref_obj, MVMnum64 value) {
+    MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
+    MVM_repr_bind_pos_multidim_n(tc, ref->body.u.multidim.obj, ref->body.u.multidim.indices, value);
+}
+void MVM_nativeref_write_multidim_s(MVMThreadContext *tc, MVMObject *ref_obj, MVMString *value) {
+    MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
+    MVM_repr_bind_pos_multidim_s(tc, ref->body.u.multidim.obj, ref->body.u.multidim.indices, value);
+}
