@@ -54,7 +54,7 @@ void MVM_args_proc_cleanup(MVMThreadContext *tc, MVMArgProcContext *ctx) {
 }
 
 MVMCallsite * MVM_args_copy_callsite(MVMThreadContext *tc, MVMArgProcContext *ctx) {
-    MVMCallsite      *res   = MVM_malloc(sizeof(MVMCallsite));
+    MVMCallsite      *res   = MVM_calloc(sizeof(MVMCallsite), 1);
     MVMCallsiteEntry *flags = NULL;
     MVMCallsiteEntry *src_flags;
     MVMint32 fsize;
@@ -76,8 +76,6 @@ MVMCallsite * MVM_args_copy_callsite(MVMThreadContext *tc, MVMArgProcContext *ct
     res->arg_flags = flags;
     res->arg_count = ctx->arg_count;
     res->num_pos   = ctx->num_pos;
-    res->has_flattening = 0;
-    res->is_interned = 0;
     return res;
 }
 
@@ -95,16 +93,13 @@ MVMCallsite * MVM_args_proc_to_callsite(MVMThreadContext *tc, MVMArgProcContext
     }
 }
 
-/* Puts the args passed to the specified frame into the current use_capture. */
 MVMObject * MVM_args_use_capture(MVMThreadContext *tc, MVMFrame *f) {
-    MVMCallCapture *capture = (MVMCallCapture *)tc->cur_usecapture;
-    if (capture->body.use_mode_frame)
-        MVM_frame_dec_ref(tc, capture->body.use_mode_frame);
-    capture->body.mode               = MVM_CALL_CAPTURE_MODE_USE;
-    capture->body.use_mode_frame     = MVM_frame_inc_ref(tc, f);
-    capture->body.apc                = &f->params;
-    capture->body.effective_callsite = MVM_args_proc_to_callsite(tc, &f->params, &capture->body.owns_callsite);
-    return tc->cur_usecapture;
+    /* We used to try and avoid some GC churn by keeping one call capture per
+     * thread that was mutated. However, its lifetime was difficult to manage,
+     * leading to leaks and subtle bugs. So, we use save_capture always now
+     * for this; we may later eliminate it using escape analysis, or treat
+     * it differently in the optimizer. */
+    return MVM_args_save_capture(tc, f);
 }
 
 MVMObject * MVM_args_save_capture(MVMThreadContext *tc, MVMFrame *frame) {
@@ -127,17 +122,6 @@ MVMObject * MVM_args_save_capture(MVMThreadContext *tc, MVMFrame *frame) {
     return cc_obj;
 }
 
-MVMCallsite * MVM_args_prepare(MVMThreadContext *tc, MVMCompUnit *cu, MVMint16 callsite_idx) {
-    /* Look up callsite. */
-    MVMCallsite * cs = cu->body.callsites[callsite_idx];
-    /* Also need to store it in cur_frame to make sure that
-     * the GC knows how to walk the args buffer, and must
-     * clear it in case we trigger GC while setting it up. */
-    tc->cur_frame->cur_args_callsite = cs;
-    memset(tc->cur_frame->args, 0, sizeof(MVMRegister) * cu->body.max_callsite_size);
-    return cs;
-}
-
 static void flatten_args(MVMThreadContext *tc, MVMArgProcContext *ctx);
 
 /* Checks that the passed arguments fall within the expected arity. */
@@ -331,8 +315,14 @@ MVMArgInfo MVM_args_get_pos_str(MVMThreadContext *tc, MVMArgProcContext *ctx, MV
     autounbox(tc, MVM_CALLSITE_ARG_STR, "string", result);
     return result;
 }
+MVMArgInfo MVM_args_get_pos_uint(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMuint32 pos, MVMuint8 required) {
+    MVMArgInfo result;
+    args_get_pos(tc, ctx, pos, required, result);
+    autounbox(tc, MVM_CALLSITE_ARG_INT, "unsigned integer", result);
+    return result;
+}
 
-#define args_get_named(tc, ctx, name, required, _type) do { \
+#define args_get_named(tc, ctx, name, required) do { \
      \
     MVMuint32 flag_pos, arg_pos; \
     result.arg.s = NULL; \
@@ -361,28 +351,34 @@ MVMArgInfo MVM_args_get_pos_str(MVMThreadContext *tc, MVMArgProcContext *ctx, MV
 
 MVMArgInfo MVM_args_get_named_obj(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMString *name, MVMuint8 required) {
     MVMArgInfo result;
-    args_get_named(tc, ctx, name, required, "object");
+    args_get_named(tc, ctx, name, required);
     autobox_switch(tc, result);
     return result;
 }
 MVMArgInfo MVM_args_get_named_int(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMString *name, MVMuint8 required) {
     MVMArgInfo result;
-    args_get_named(tc, ctx, name, required, "integer");
+    args_get_named(tc, ctx, name, required);
     autounbox(tc, MVM_CALLSITE_ARG_INT, "integer", result);
     return result;
 }
 MVMArgInfo MVM_args_get_named_num(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMString *name, MVMuint8 required) {
     MVMArgInfo result;
-    args_get_named(tc, ctx, name, required, "number");
+    args_get_named(tc, ctx, name, required);
     autounbox(tc, MVM_CALLSITE_ARG_NUM, "number", result);
     return result;
 }
 MVMArgInfo MVM_args_get_named_str(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMString *name, MVMuint8 required) {
     MVMArgInfo result;
-    args_get_named(tc, ctx, name, required, "string");
+    args_get_named(tc, ctx, name, required);
     autounbox(tc, MVM_CALLSITE_ARG_STR, "string", result);
     return result;
 }
+MVMArgInfo MVM_args_get_named_uint(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMString *name, MVMuint8 required) {
+    MVMArgInfo result;
+    args_get_named(tc, ctx, name, required);
+    autounbox(tc, MVM_CALLSITE_ARG_INT, "unsigned integer", result);
+    return result;
+}
 MVMint64 MVM_args_has_named(MVMThreadContext *tc, MVMArgProcContext *ctx, MVMString *name) {
     MVMuint32 flag_pos, arg_pos;
     for (flag_pos = arg_pos = ctx->num_pos; arg_pos < ctx->arg_count; flag_pos++, arg_pos += 2)
@@ -460,10 +456,9 @@ void MVM_args_set_result_int(MVMThreadContext *tc, MVMint64 result, MVMint32 fra
             case MVM_RETURN_NUM:
                 target->return_value->n64 = (MVMnum64)result;
                 break;
-            case MVM_RETURN_OBJ: {
+            case MVM_RETURN_OBJ:
                 autobox(tc, target, result, int_box_type, 0, set_int, target->return_value->o);
                 break;
-            }
             default:
                 MVM_exception_throw_adhoc(tc, "Result return coercion from int NYI; expects type %u", target->return_type);
         }
@@ -481,10 +476,9 @@ void MVM_args_set_result_num(MVMThreadContext *tc, MVMnum64 result, MVMint32 fra
             case MVM_RETURN_INT:
                 target->return_value->i64 = (MVMint64)result;
                 break;
-            case MVM_RETURN_OBJ: {
+            case MVM_RETURN_OBJ:
                 autobox(tc, target, result, num_box_type, 0, set_num, target->return_value->o);
                 break;
-            }
             default:
                 MVM_exception_throw_adhoc(tc, "Result return coercion from num NYI; expects type %u", target->return_type);
         }
@@ -499,10 +493,9 @@ void MVM_args_set_result_str(MVMThreadContext *tc, MVMString *result, MVMint32 f
             case MVM_RETURN_STR:
                 target->return_value->s = result;
                 break;
-            case MVM_RETURN_OBJ: {
+            case MVM_RETURN_OBJ:
                 autobox(tc, target, result, str_box_type, 1, set_str, target->return_value->o);
                 break;
-            }
             default:
                 MVM_exception_throw_adhoc(tc, "Result return coercion from str NYI; expects type %u", target->return_type);
         }
@@ -839,7 +832,7 @@ void MVM_args_setup_thunk(MVMThreadContext *tc, MVMRegister *res_reg, MVMReturnT
     cur_frame->return_value      = res_reg;
     cur_frame->return_type       = return_type;
     cur_frame->return_address    = *(tc->interp_cur_op);
-    cur_frame->cur_args_callsite = callsite; 
+    cur_frame->cur_args_callsite = callsite;
 }
 
 /* Custom bind failure handling. Invokes the HLL's bind failure handler, with
