@@ -37,7 +37,7 @@ void MVM_spesh_use_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshFacts *
 
 /* Obtains a string constant. */
 MVMString * MVM_spesh_get_string(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshOperand o) {
-    return g->sf->body.cu->body.strings[o.lit_str_idx];
+    return MVM_cu_string(tc, g->sf->body.cu, o.lit_str_idx);
 }
 
 /* Copy facts between two register operands. */
@@ -872,6 +872,10 @@ static void optimize_istrue_isfalse(MVMThreadContext *tc, MVMSpeshGraph *g, MVMS
 
         switch (bs == NULL ? MVM_BOOL_MODE_NOT_TYPE_OBJECT : bs->mode) {
             case MVM_BOOL_MODE_UNBOX_INT:
+                /* This optimization can only handle values known to be concrete. */
+                if (!(facts->flags & MVM_SPESH_FACT_CONCRETE)) {
+                    return;
+                }
                 /* We can just unbox the int and pretend it's a bool. */
                 ins->info = MVM_op_get_op(MVM_OP_unbox_i);
                 if (negated_op)
@@ -1093,6 +1097,17 @@ static void optimize_call(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb
                         }
                     }
                 }
+                else if (!MVM_is_null(tc, is->class_handle)) {
+                    /* This type of code object supports multi-dispatch,
+                     * but we actually have a single dispatch routine. */
+                    MVMRegister dest;
+                    REPR(code)->attr_funcs.get_attribute(tc,
+                        STABLE(code), code, OBJECT_BODY(code),
+                        is->class_handle, is->attr_name,
+                        is->hint, &dest, MVM_reg_obj);
+                    if (REPR(dest.o)->ID == MVM_REPR_ID_MVMCode)
+                        target = dest.o;
+                }
             }
             else if (!MVM_is_null(tc, is->class_handle)) {
                 /* Single dispatch; retrieve the code object. */
@@ -1509,6 +1524,8 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
         case MVM_OP_bindattrs_n:
         case MVM_OP_bindattrs_s:
         case MVM_OP_bindattrs_o:
+        case MVM_OP_assign_i:
+        case MVM_OP_assign_n:
             optimize_repr_op(tc, g, bb, ins, 0);
             break;
         case MVM_OP_atpos_i:
@@ -1538,6 +1555,10 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
         case MVM_OP_getattrs_n:
         case MVM_OP_getattrs_s:
         case MVM_OP_getattrs_o:
+        case MVM_OP_decont_i:
+        case MVM_OP_decont_n:
+        case MVM_OP_decont_s:
+        case MVM_OP_decont_u:
         case MVM_OP_create:
             optimize_repr_op(tc, g, bb, ins, 1);
             break;
