@@ -70,28 +70,41 @@ static MVMString * re_nfg(MVMThreadContext *tc, MVMString *in) {
     MVMCodepointIter ci;
     MVMint32 ready;
     MVMString *out;
+    MVMuint32 bufsize = in->body.num_graphs;
 
-    /* Create output buffer; it'll never be longer than the initial estimate
-     * since the most we'll do is collapse two things into one in places. */
-    MVMGrapheme32 *out_buffer = MVM_malloc(in->body.num_graphs * sizeof(MVMGrapheme32));
+    /* Create the output buffer. We used to believe it can't ever be bigger
+     * than the initial estimate, but utf8-c8 showed us otherwise. */
+    MVMGrapheme32 *out_buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
     MVMint64 out_pos = 0;
 
     /* Iterate codepoints and normalizer. */
     MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
-    MVM_string_ci_init(tc, &ci, in);
+    MVM_string_ci_init(tc, &ci, in, 0);
     while (MVM_string_ci_has_more(tc, &ci)) {
         MVMGrapheme32 g;
         ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, MVM_string_ci_get_codepoint(tc, &ci), &g);
         if (ready) {
+            if (out_pos + ready > bufsize) {
+                /* Doubling up the buffer size seems excessive, so just
+                 * add a generous amount of storage */
+                bufsize += ready + 32;
+                out_buffer = MVM_realloc(out_buffer, bufsize * sizeof(MVMGrapheme32));
+            }
             out_buffer[out_pos++] = g;
-            while (--ready > 0)
+            while (--ready > 0) {
                 out_buffer[out_pos++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+            }
         }
     }
     MVM_unicode_normalizer_eof(tc, &norm);
     ready = MVM_unicode_normalizer_available(tc, &norm);
-    while (ready--)
+    if (out_pos + ready > bufsize) {
+        bufsize += ready + 1;
+        out_buffer = MVM_realloc(out_buffer, bufsize * sizeof(MVMGrapheme32));
+    }
+    while (ready--) {
         out_buffer[out_pos++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+    }
     MVM_unicode_normalizer_cleanup(tc, &norm);
 
     /* Build result string. */
@@ -402,14 +415,16 @@ MVMString * MVM_string_concatenate(MVMThreadContext *tc, MVMString *a, MVMString
             MVMString *effective_a = a;
             MVMString *effective_b = b;
             if (strands_a + strands_b > MVM_STRING_MAX_STRANDS) {
-                if (strands_a >= strands_b) {
-                    effective_a = collapse_strands(tc, effective_a);
-                    strands_a   = 1;
-                }
-                else {
-                    effective_b = collapse_strands(tc, effective_b);
-                    strands_b   = 1;
-                }
+                MVMROOT(tc, result, {
+                    if (strands_a >= strands_b) {
+                        effective_a = collapse_strands(tc, effective_a);
+                        strands_a   = 1;
+                    }
+                    else {
+                        effective_b = collapse_strands(tc, effective_b);
+                        strands_b   = 1;
+                    }
+                });
             }
 
             /* Assemble the result. */
@@ -539,15 +554,47 @@ MVMint64 MVM_string_equal_at_ignore_case(MVMThreadContext *tc, MVMString *a, MVM
     return MVM_string_equal_at(tc, lca, lcb, offset);
 }
 
+MVMGrapheme32 MVM_string_ord_at(MVMThreadContext *tc, MVMString *s, MVMint64 offset) {
+    MVMStringIndex agraphs;
+    MVMGrapheme32 g;
+
+    MVM_string_check_arg(tc, s, "grapheme_at");
+
+    agraphs = MVM_string_graphs(tc, s);
+    if (offset < 0 || offset >= agraphs)
+	return -1;
+
+    g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
+
+    return g >= 0 ? g : MVM_nfg_get_synthetic_info(tc, g)->base;
+}
+
 MVMGrapheme32 MVM_string_ord_basechar_at(MVMThreadContext *tc, MVMString *s, MVMint64 offset) {
-    MVMGrapheme32 g = MVM_string_get_grapheme_at(tc, s, offset);
+    MVMStringIndex agraphs;
+    MVMGrapheme32 g;
     MVMNormalizer norm;
     MVMint32 ready;
-    MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
-    ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
-    MVM_unicode_normalizer_eof(tc, &norm);
-    if (!ready)
-        g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+
+    MVM_string_check_arg(tc, s, "grapheme_at");
+
+    agraphs = MVM_string_graphs(tc, s);
+    if (offset < 0 || offset >= agraphs)
+	return -1;  /* fixes RT #126771 */
+
+    g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
+
+    if (g < 0) {
+	MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
+	g = si->base;
+    }
+    else {
+	MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
+	ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
+	MVM_unicode_normalizer_eof(tc, &norm);
+	if (!ready)
+	    g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+    }
+
     return g;
 }
 
@@ -780,18 +827,20 @@ MVMString * MVM_string_decode(MVMThreadContext *tc,
 }
 
 /* Encodes an MVMString to a C buffer, dependent on the encoding type flag */
-char * MVM_string_encode(MVMThreadContext *tc, MVMString *s, MVMint64 start, MVMint64 length, MVMuint64 *output_size, MVMint64 encoding_flag, MVMString *replacement) {
+char * MVM_string_encode(MVMThreadContext *tc, MVMString *s, MVMint64 start,
+        MVMint64 length, MVMuint64 *output_size, MVMint64 encoding_flag,
+        MVMString *replacement, MVMint32 translate_newlines) {
     switch(encoding_flag) {
         case MVM_encoding_type_utf8:
-            return MVM_string_utf8_encode_substr(tc, s, output_size, start, length, replacement);
+            return MVM_string_utf8_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_ascii:
-            return MVM_string_ascii_encode_substr(tc, s, output_size, start, length, replacement);
+            return MVM_string_ascii_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_latin1:
-            return MVM_string_latin1_encode_substr(tc, s, output_size, start, length, replacement);
+            return MVM_string_latin1_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_utf16:
-            return MVM_string_utf16_encode_substr(tc, s, output_size, start, length, replacement);
+            return MVM_string_utf16_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_windows1252:
-            return MVM_string_windows1252_encode_substr(tc, s, output_size, start, length, replacement);
+            return MVM_string_windows1252_encode_substr(tc, s, output_size, start, length, replacement, translate_newlines);
         case MVM_encoding_type_utf8_c8:
             return MVM_string_utf8_c8_encode_substr(tc, s, output_size, start, length, replacement);
         default:
@@ -835,7 +884,7 @@ void MVM_string_encode_to_buf(MVMThreadContext *tc, MVMString *s, MVMString *enc
     MVMROOT(tc, s, {
         const MVMuint8 encoding_flag = MVM_string_find_encoding(tc, enc_name);
         encoded = (MVMuint8 *)MVM_string_encode(tc, s, 0, MVM_string_graphs(tc, s), &output_size,
-            encoding_flag, replacement);
+            encoding_flag, replacement, 0);
     });
     });
 
@@ -1669,17 +1718,17 @@ MVMuint8 MVM_string_find_encoding(MVMThreadContext *tc, MVMString *name) {
     if (!encoding_name_init) {
         MVM_gc_allocate_gen2_default_set(tc);
         encoding_utf8_name        = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf8");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf8_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_utf8_name, "Encoding name");
         encoding_ascii_name       = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "ascii");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_ascii_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_ascii_name, "Encoding name");
         encoding_latin1_name      = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "iso-8859-1");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_latin1_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_latin1_name, "Encoding name");
         encoding_utf16_name       = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf16");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf16_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_utf16_name, "Encoding name");
         encoding_windows1252_name = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "windows-1252");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_windows1252_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_windows1252_name, "Encoding name");
         encoding_utf8_c8_name     = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf8-c8");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf8_c8_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_utf8_c8_name, "Encoding name");
         encoding_name_init   = 1;
         MVM_gc_allocate_gen2_default_clear(tc);
     }
